@@ -316,30 +316,42 @@ static int do_sasl_interact(LDAP UNUSED(*ld),unsigned UNUSED(flags),void *defaul
       case SASL_CB_GETREALM:
         if (cfg->ldc_sasl_realm)
         {
+          log_log(LOG_DEBUG,"do_sasl_interact(): returning sasl_realm \"%s\"",cfg->ldc_sasl_realm);
           interact->result=cfg->ldc_sasl_realm;
           interact->len=strlen(cfg->ldc_sasl_realm);
         }
+        else
+          log_log(LOG_DEBUG,"do_sasl_interact(): were asked for sasl_realm but we don't have any");
         break;
       case SASL_CB_AUTHNAME:
         if (cfg->ldc_sasl_authcid)
         {
+          log_log(LOG_DEBUG,"do_sasl_interact(): returning sasl_authcid \"%s\"",cfg->ldc_sasl_authcid);
           interact->result=cfg->ldc_sasl_authcid;
           interact->len=strlen(cfg->ldc_sasl_authcid);
         }
+        else
+          log_log(LOG_DEBUG,"do_sasl_interact(): were asked for sasl_authcid but we don't have any");
         break;
       case SASL_CB_USER:
         if (cfg->ldc_sasl_authzid)
         {
+          log_log(LOG_DEBUG,"do_sasl_interact(): returning sasl_authzid \"%s\"",cfg->ldc_sasl_authzid);
           interact->result=cfg->ldc_sasl_authzid;
           interact->len=strlen(cfg->ldc_sasl_authzid);
         }
+        else
+          log_log(LOG_DEBUG,"do_sasl_interact(): were asked for sasl_authzid but we don't have any");
         break;
       case SASL_CB_PASS:
         if (cfg->ldc_bindpw)
         {
+          log_log(LOG_DEBUG,"do_sasl_interact(): returning bindpw \"***\"");
           interact->result=cfg->ldc_bindpw;
           interact->len=strlen(cfg->ldc_bindpw);
         }
+        else
+          log_log(LOG_DEBUG,"do_sasl_interact(): were asked for bindpw but we don't have any");
         break;
       default:
         /* just ignore */
@@ -395,35 +407,27 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
                       (session->bindpw[0]!='\0')?"\"***\"":"\"\"",uri);
     return ldap_simple_bind_s(session->ld,session->binddn,session->bindpw);
   }
+  /* perform SASL bind if requested and available on platform */
 #ifdef HAVE_LDAP_SASL_INTERACTIVE_BIND_S
   /* TODO: store this information in the session */
-  if (!nslcd_cfg->ldc_usesasl)
-  {
-#endif /* HAVE_LDAP_SASL_INTERACTIVE_BIND_S */
-    /* do a simple bind */
-    if (nslcd_cfg->ldc_binddn)
-      log_log(LOG_DEBUG,"ldap_simple_bind_s(\"%s\",%s) (uri=\"%s\")",nslcd_cfg->ldc_binddn,
-                        nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
-    else
-      log_log(LOG_DEBUG,"ldap_simple_bind_s(NULL,%s) (uri=\"%s\")",
-                        nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
-    return ldap_simple_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_bindpw);
-#ifdef HAVE_LDAP_SASL_INTERACTIVE_BIND_S
-  }
-  else
+  if (nslcd_cfg->ldc_sasl_mech!=NULL)
   {
     /* do a SASL bind */
-    log_log(LOG_DEBUG,"SASL bind to %s as %s",uri,nslcd_cfg->ldc_binddn);
     if (nslcd_cfg->ldc_sasl_secprops!=NULL)
     {
       log_log(LOG_DEBUG,"ldap_set_option(LDAP_OPT_X_SASL_SECPROPS,\"%s\")",nslcd_cfg->ldc_sasl_secprops);
       LDAP_SET_OPTION(session->ld,LDAP_OPT_X_SASL_SECPROPS,(void *)nslcd_cfg->ldc_sasl_secprops);
     }
 #ifdef HAVE_SASL_INTERACT_T
-    log_log(LOG_DEBUG,"ldap_sasl_interactive_bind_s(\"%s\",\"%s\")",nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech);
+    if (nslcd_cfg->ldc_binddn!=NULL)
+      log_log(LOG_DEBUG,"ldap_sasl_interactive_bind_s(\"%s\",\"%s\") (uri=\"%s\")",
+            nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech,uri);
+    else
+      log_log(LOG_DEBUG,"ldap_sasl_interactive_bind_s(NULL,\"%s\") (uri=\"%s\")",
+            nslcd_cfg->ldc_sasl_mech,uri);
     return ldap_sasl_interactive_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech,NULL,NULL,
-                                    LDAP_SASL_QUIET,
-                                    do_sasl_interact,(void *)nslcd_cfg);
+                                        LDAP_SASL_QUIET,
+                                        do_sasl_interact,(void *)nslcd_cfg);
 #else /* HAVE_SASL_INTERACT_T */
     if (nslcd_cfg->ldc_bindpw!=NULL)
     {
@@ -435,12 +439,26 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
       cred.bv_val="";
       cred.bv_len=0;
     }
-    log_log(LOG_DEBUG,"ldap_sasl_bind_s(\"%s\",\"%s\",%s)",nslcd_cfg->ldc_binddn,
-                      nslcd_cfg->ldc_sasl_mech,nslcd_cfg->ldc_bindpw?"\"***\"":"NULL");
+    if (nslcd_cfg->ldc_binddn!=NULL)
+      log_log(LOG_DEBUG,"ldap_sasl_bind_s(\"%s\",\"%s\",%s) (uri=\"%s\")",
+            nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech,
+            nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
+    else
+      log_log(LOG_DEBUG,"ldap_sasl_bind_s(NULL,\"%s\",%s) (uri=\"%s\")",
+            nslcd_cfg->ldc_sasl_mech,
+            nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
     return ldap_sasl_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech,&cred,NULL,NULL,NULL);
 #endif /* not HAVE_SASL_INTERACT_T */
   }
 #endif /* HAVE_LDAP_SASL_INTERACTIVE_BIND_S */
+  /* do a simple bind */
+  if (nslcd_cfg->ldc_binddn)
+    log_log(LOG_DEBUG,"ldap_simple_bind_s(\"%s\",%s) (uri=\"%s\")",nslcd_cfg->ldc_binddn,
+                      nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
+  else
+    log_log(LOG_DEBUG,"ldap_simple_bind_s(NULL,%s) (uri=\"%s\")",
+                      nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
+  return ldap_simple_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_bindpw);
 }
 
 #ifdef HAVE_LDAP_SET_REBIND_PROC
@@ -529,6 +547,20 @@ static void do_close(MYLDAP_SESSION *session)
 {
   int i;
   int rc;
+  int sd=-1;
+  struct timeval tv;
+  /* set timeout options on socket to avoid hang in some cases
+     (we set a short timeout because we don't care too much about properly
+     shutting down the connection) */
+  if (ldap_get_option(session->ld,LDAP_OPT_DESC,&sd)==LDAP_SUCCESS)
+  {
+    /* ignore errors */
+    tv.tv_sec=nslcd_cfg->ldc_timelimit/2;
+    if (!tv.tv_sec) tv.tv_sec=1;
+    tv.tv_usec=0;
+    (void)setsockopt(sd,SOL_SOCKET,SO_RCVTIMEO,(void *)&tv,sizeof(tv));
+    (void)setsockopt(sd,SOL_SOCKET,SO_SNDTIMEO,(void *)&tv,sizeof(tv));
+  }
   /* if we had reachability problems with the server close the connection */
   if (session->ld!=NULL)
   {
@@ -603,7 +635,7 @@ static void myldap_session_check(MYLDAP_SESSION *session)
    and binds to the server. This returns an LDAP status code. */
 static int do_open(MYLDAP_SESSION *session)
 {
-  int rc,rc2;
+  int rc;
   int sd=-1;
   int off=0;
   /* check if the idle time for the connection has expired */
@@ -626,13 +658,7 @@ static int do_open(MYLDAP_SESSION *session)
                         ldap_err2string(rc),(errno==0)?"":": ",
                         (errno==0)?"":strerror(errno));
     if (session->ld!=NULL)
-    {
-      log_log(LOG_DEBUG,"ldap_unbind()");
-      rc2=ldap_unbind(session->ld);
-      session->ld=NULL;
-      if (rc2!=LDAP_SUCCESS)
-        log_log(LOG_WARNING,"ldap_unbind() failed: %s",ldap_err2string(rc2));
-    }
+      do_close(session);
     return rc;
   }
   else if (session->ld==NULL)
@@ -644,10 +670,7 @@ static int do_open(MYLDAP_SESSION *session)
   rc=do_set_options(session);
   if (rc!=LDAP_SUCCESS)
   {
-    rc2=ldap_unbind(session->ld);
-    session->ld=NULL;
-    if (rc2!=LDAP_SUCCESS)
-      log_log(LOG_WARNING,"ldap_unbind() failed: %s",ldap_err2string(rc2));
+    do_close(session);
     return rc;
   }
   /* bind to the server */
@@ -661,10 +684,7 @@ static int do_open(MYLDAP_SESSION *session)
                         nslcd_cfg->ldc_uris[session->current_uri].uri,
                         ldap_err2string(rc),(errno==0)?"":": ",
                         (errno==0)?"":strerror(errno));
-    rc2=ldap_unbind(session->ld);
-    session->ld=NULL;
-    if (rc2!=LDAP_SUCCESS)
-      log_log(LOG_WARNING,"ldap_unbind() failed: %s",ldap_err2string(rc2));
+    do_close(session);
     return rc;
   }
   /* disable keepalive on the LDAP connection socket (why?) */
@@ -1042,36 +1062,6 @@ MYLDAP_ENTRY *myldap_get_entry(MYLDAP_SEARCH *search,int *rcp)
     /* handle result */
     switch (rc)
     {
-      case -1:
-        /* we have an error condition, try to get error code */
-        if (ldap_get_option(search->session->ld,LDAP_OPT_ERROR_NUMBER,&rc)!=LDAP_SUCCESS)
-          rc=LDAP_UNAVAILABLE;
-        log_log(LOG_ERR,"ldap_result() failed: %s",ldap_err2string(rc));
-        /* close connection on connection problems */
-        if ((rc==LDAP_UNAVAILABLE)||(rc==LDAP_SERVER_DOWN)||(rc==LDAP_SUCCESS))
-        {
-          /* close the connection and retry */
-          do_close(search->session);
-          if (search->may_retry_search)
-          {
-            log_log(LOG_DEBUG,"myldap_get_entry(): retry search");
-            search->may_retry_search=0;
-            if (do_retry_search(search)==LDAP_SUCCESS)
-              return myldap_get_entry(search,rcp);
-          }
-        }
-        /* close search */
-        myldap_search_close(search);
-        if (rcp!=NULL)
-          *rcp=rc;
-        return NULL;
-      case 0:
-        /* the timeout expired */
-        log_log(LOG_ERR,"ldap_result() timed out");
-        myldap_search_close(search);
-        if (rcp!=NULL)
-          *rcp=LDAP_TIMELIMIT_EXCEEDED;
-        return NULL;
       case LDAP_RES_SEARCH_ENTRY:
         /* we have a normal search entry, update timestamp and return result */
         time(&(search->session->lastactivity));
@@ -1187,11 +1177,46 @@ MYLDAP_ENTRY *myldap_get_entry(MYLDAP_SEARCH *search,int *rcp)
       case LDAP_RES_SEARCH_REFERENCE:
         break; /* just ignore search references */
       default:
-        log_log(LOG_WARNING,"ldap_result() returned unexpected result type");
+        /* we have some error condition, find out which */
+        switch (rc)
+        {
+          case -1:
+            /* try to get error code */
+            if (ldap_get_option(search->session->ld,LDAP_OPT_ERROR_NUMBER,&rc)!=LDAP_SUCCESS)
+              rc=LDAP_UNAVAILABLE;
+            log_log(LOG_ERR,"ldap_result() failed: %s",ldap_err2string(rc));
+            break;
+          case 0:
+            /* the timeout expired */
+            log_log(LOG_ERR,"ldap_result() timed out");
+            rc=LDAP_TIMELIMIT_EXCEEDED;
+            break;
+          default:
+            /* unknown code */
+            log_log(LOG_WARNING,"ldap_result() returned unexpected result type");
+            rc=LDAP_PROTOCOL_ERROR;
+        }
+        /* close connection on some connection problems */
+        if ((rc==LDAP_UNAVAILABLE)||(rc==LDAP_SERVER_DOWN)||(rc==LDAP_SUCCESS)||
+            (rc==LDAP_TIMELIMIT_EXCEEDED)|(rc==LDAP_OPERATIONS_ERROR)||
+            (rc==LDAP_PROTOCOL_ERROR))
+        {
+          do_close(search->session);
+          /* retry once if no data has been received yet */
+          if (search->may_retry_search)
+          {
+            log_log(LOG_DEBUG,"myldap_get_entry(): retry search");
+            search->may_retry_search=0;
+            if (do_retry_search(search)==LDAP_SUCCESS)
+              return myldap_get_entry(search,rcp);
+          }
+        }
+        /* close search */
         myldap_search_close(search);
         if (rcp!=NULL)
-          *rcp=LDAP_PROTOCOL_ERROR;
+          *rcp=rc;
         return NULL;
+
     }
   }
 }
