@@ -148,8 +148,8 @@ int nslcd_pam_authc(TFILE *fp,MYLDAP_SESSION *session)
   /* write the response header */
   WRITE_INT32(fp,NSLCD_VERSION);
   WRITE_INT32(fp,NSLCD_ACTION_PAM_AUTHC);
-  /* if the username is blank and rootpwmoddn is configure, try to authenticate
-     as administrator, otherwise validate request as usual */
+  /* if the username is blank and rootpwmoddn is configured, try to
+     authenticate as administrator, otherwise validate request as usual */
   if ((*username=='\0')&&(nslcd_cfg->ldc_rootpwmoddn!=NULL))
   {
     if (strlen(nslcd_cfg->ldc_rootpwmoddn)>=sizeof(userdn))
@@ -166,6 +166,8 @@ int nslcd_pam_authc(TFILE *fp,MYLDAP_SESSION *session)
   }
   /* try authentication */
   rc=try_bind(userdn,password);
+  if (rc==NSLCD_PAM_SUCCESS)
+    log_log(LOG_DEBUG,"bind successful");
   /* write response */
   WRITE_INT32(fp,NSLCD_RESULT_BEGIN);
   WRITE_STRING(fp,username);
@@ -272,7 +274,7 @@ int nslcd_pam_authz(TFILE *fp,MYLDAP_SESSION *session)
   char username[256];
   char userdn[256];
   char servicename[64];
-  char ruser[32];
+  char ruser[256];
   char rhost[256];
   char tty[256];
   char hostname[HOST_NAME_MAX+1];
